from django.apps import apps as installed_apps
from django.db import connections, router
from tqdm import tqdm

from .crypto import Crypto


class Migrator:
    def __init__(self, app, model, field, key=None, tofield=None, idfield='id'):
        self.app = app
        self.model = model.lower()
        self.field = field.lower()
        self.crypto = Crypto(key)
        self.tofield = tofield
        self.idfield = idfield

    def encrypt(self, apps=None, schema_editor=None, offset=0, total=None, limit=1000):
        return self.executor(apps, schema_editor, offset, total, limit, method='encrypt')

    def decrypt(self, apps=None, schema_editor=None, offset=0, total=None, limit=1000):
        return self.executor(apps, schema_editor, offset, total, limit, method='decrypt')

    def copy_to(self, apps=None, schema_editor=None, offset=0, total=None, limit=1000):
        return self.executor(apps, schema_editor, offset, total, limit, method='copy_to')

    def encrypt_to(self, apps=None, schema_editor=None, offset=0, total=None, limit=1000):
        return self.executor(apps, schema_editor, offset, total, limit, method='encrypt_to')

    def decrypt_to(self, apps=None, schema_editor=None, offset=0, total=None, limit=1000):
        return self.executor(apps, schema_editor, offset, total, limit, method='decrypt_to')

    def executor(self, apps=None, schema_editor=None, offset=0, total=None, limit=1000, method=None):
        if not method:
            return

        if not apps:
            apps = installed_apps
        model = apps.get_model(self.app, self.model)
        if not schema_editor:
            db_alias = router.db_for_write(model=model)
        else:
            db_alias = schema_editor.connection.alias
        db_table = model._meta.db_table if model._meta.db_table else f"{self.app}_{self.model}"
        if limit == -1:
            total = model.objects.using(db_alias).count()
        else:
            if not total:
                last_record = model.objects.using(db_alias).order_by(f"-{self.idfield}").first()
                if last_record:
                    total = getattr(last_record, self.idfield)
                else:
                    total = 0
            if limit > total:
                limit = total

        t = tqdm(total=total-offset)
        while offset < total:
            value_list = []
            with connections[db_alias].cursor() as cursor:
                if limit == -1:
                    cursor.execute(f"select {self.idfield}, {self.field} from {db_table};")
                else:
                    cursor.execute(
                        f"select {self.idfield}, {self.field} from {db_table} where "
                        f"{self.idfield}>{offset} order by {self.idfield} limit {limit};"
                    )
                for query in cursor.fetchall():
                    if method in ['encrypt', 'encrypt_to']:
                        value_list.append([query[0], self.crypto.encrypt(query[1])])
                    elif method in ['decrypt', 'decrypt_to']:
                        text = self.crypto.decrypt(query[1]) or ''
                        value_list.append([query[0], text.replace("'", "''")])
                    elif method == 'copy_to':
                        text = query[1] or ''
                        value_list.append([query[0], text.replace("'", "''")])
                execute_sql = ''
                for value in value_list:
                    if method in ['encrypt', 'decrypt']:
                        if value[1] is None:
                            execute_sql += (
                                f"update {db_table} set {self.field}=NULL where {self.idfield}='{value[0]}';"
                            )
                        else:
                            execute_sql += (
                                f"update {db_table} set {self.field}='{value[1]}' where {self.idfield}='{value[0]}';"
                            )
                    elif method in ['copy_to', 'encrypt_to', 'decrypt_to']:
                        if value[1] is None:
                            execute_sql += (
                                f"update {db_table} set {self.tofield}=NULL where {self.idfield}='{value[0]}';"
                            )
                        else:
                            execute_sql += (
                                f"update {db_table} set {self.tofield}='{value[1]}' where {self.idfield}='{value[0]}';"
                            )
                cursor.execute(execute_sql)
            if value_list:
                if limit == -1:
                    t.update(len(value_list) - offset)
                    offset = total
                else:
                    t.update(value_list[-1][0] - offset)
                    offset = value_list[-1][0]
        t.close()



from django.core.management.base import BaseCommand

from mirage.tools import Migrator


class Command(BaseCommand):
    help = "encrypt/decrypt data use Mirage ~"

    def add_arguments(self, parser):
        parser.add_argument("--app", type=str, required=True)
        parser.add_argument("--model", type=str, required=True)
        parser.add_argument("--field", type=str, required=True)
        parser.add_argument("--method", type=str, required=True)
        parser.add_argument("--offset", type=int, required=False)
        parser.add_argument("--total", type=int, required=False)
        parser.add_argument("--limit", type=int, required=False)
        parser.add_argument("--tofield", type=str, required=False)
        parser.add_argument("--idfield", type=str, required=False)

    def handle(self, *args, **options):
        app = options["app"]
        model = options["model"]
        field = options["field"]
        method = options["method"]
        offset = options["offset"] or 0
        total = options["total"]
        limit = options["limit"] or 1000
        tofield = options["tofield"]
        idfield = options["idfield"] or "id"

        if method == "encrypt":
            Migrator(app, model, field, idfield=idfield).encrypt(
                offset=offset, total=total, limit=limit
            )
        elif method == "decrypt":
            Migrator(app, model, field, idfield=idfield).decrypt(
                offset=offset, total=total, limit=limit
            )
        elif method == "copy_to":
            assert tofield is not None
            Migrator(app, model, field, tofield=tofield, idfield=idfield).copy_to(
                offset=offset, total=total, limit=limit
            )
        elif method == "decrypt_to":
            assert tofield is not None
            Migrator(app, model, field, tofield=tofield, idfield=idfield).decrypt_to(
                offset=offset, total=total, limit=limit
            )
        elif method == "encrypt_to":
            assert tofield is not None
            Migrator(app, model, field, tofield=tofield, idfield=idfield).encrypt_to(
                offset=offset, total=total, limit=limit
            )
        else:
            return



import base64

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from django.conf import settings
from django.utils.encoding import force_bytes, force_str


class BaseCipher:
    def __init__(self, key, iv) -> None:
        self.key = key
        self.iv = iv

    def encrypt(self, text) -> str:
        return text

    def decrypt(self, encrypted) -> str:
        return encrypted


class ECBCipher(BaseCipher):

    def encrypt(self, text):
        encryptor = Cipher(algorithms.AES(self.key), modes.ECB(), default_backend()).encryptor()
        padder = padding.PKCS7(algorithms.AES(self.key).block_size).padder()
        padded_data = padder.update(force_bytes(text)) + padder.finalize()
        encrypted_text = encryptor.update(padded_data) + encryptor.finalize()
        return force_str(base64.urlsafe_b64encode(encrypted_text))

    def decrypt(self, encrypted):
        decryptor = Cipher(algorithms.AES(self.key), modes.ECB(), default_backend()).decryptor()
        padder = padding.PKCS7(algorithms.AES(self.key).block_size).unpadder()
        decrypted_text = decryptor.update(base64.urlsafe_b64decode(encrypted))
        unpadded_text = padder.update(decrypted_text) + padder.finalize()
        return force_str(unpadded_text)

class CBCCipher(BaseCipher):

    def encrypt(self, text) -> str:
        encryptor = Cipher(algorithms.AES(self.key), modes.CBC(self.iv), default_backend()).encryptor()
        padder = padding.PKCS7(algorithms.AES(self.key).block_size).padder()
        padded_data = padder.update(force_bytes(text)) + padder.finalize()
        encrypted_text = encryptor.update(padded_data) + encryptor.finalize()
        return force_str(base64.urlsafe_b64encode(encrypted_text))

    def decrypt(self, encrypted) -> str:
        decryptor = Cipher(algorithms.AES(self.key), modes.CBC(self.iv), default_backend()).decryptor()
        padder = padding.PKCS7(algorithms.AES(self.key).block_size).unpadder()
        decrypted_text = decryptor.update(base64.urlsafe_b64decode(encrypted))
        unpadded_text = padder.update(decrypted_text) + padder.finalize()
        return force_str(unpadded_text)

class Crypto:

    def __init__(self, key=None, mode=None, iv=None):
        if key is None:
            key = getattr(settings, "MIRAGE_SECRET_KEY", None) or getattr(settings, "SECRET_KEY")
        assert len(key) >= 32, "mirage key length must more than 32!"
        key = base64.urlsafe_b64encode(force_bytes(key))[:32]
        if mode is None:
            mode = getattr(settings, "MIRAGE_CIPHER_MODE", "ECB")
        if iv is None:
            iv=getattr(settings, "MIRAGE_CIPHER_IV", "1234567890abcdef")
        self.cipher = eval(f"{mode}Cipher")(key=key, iv=force_bytes(iv))

    def encrypt(self, text):
        if text is None:
            return None
        try:
            self.cipher.decrypt(text)
            return text
        except Exception:
            return self.cipher.encrypt(text)

    def decrypt(self, encrypted):
        if encrypted is None:
            return None
        try:
            return self.cipher.decrypt(encrypted)
        except Exception:
            return encrypted



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf8') as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))
setup(
    name='django-mirage-field',
    version='1.2.2',
    install_requires=[
        "cryptography",
        "tqdm",
    ],
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    license='MIT License',
    description='A Django model field that encrypt your data when save to and decrypt when get from database. It keeps data always encrypted in database.',
    long_description_content_type="text/markdown",
    long_description=README,
    url='https://github.com/luojilab/django-mirage-field',
    author='tcitry',
    author_email='tcitry@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
    ],
)
